from unittest.mock import NonCallableMagicMock
import numpy as np
import math
import torch
from tqdm import tqdm
import warnings

SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))


def get_ESS(logw):
    logw_max = np.max(logw)
    logw_normed = logw - logw_max

    weights = np.exp(logw_normed) / np.sum(np.exp(logw_normed))
    return 1.0 / np.sum(weights * weights) / len(weights)


def resample_equal(samples, weights, rstate=None):
    """
        Resample a new set of points from the weighted set of inputs
        such that they all have equal weight.
        Each input sample appears in the output array either
        `floor(weights[i] * nsamples)` or `ceil(weights[i] * nsamples)` times,
        with `floor` or `ceil` randomly selected (weighted by proximity).

    Parameters
    ----------
    samples : `~numpy.ndarray` with shape (nsamples,)
        Set of unequally weighted samples.
    weights : `~numpy.ndarray` with shape (nsamples,)
        Corresponding weight of each sample.
    rstate : `~numpy.random.RandomState`, optional
        `~numpy.random.RandomState` instance.

    Returns
    -------
    equal_weight_samples : `~numpy.ndarray` with shape (nsamples,)
        New set of samples with equal weights.
    
    Examples
    --------
    >>> x = np.array([[1., 1.], [2., 2.], [3., 3.], [4., 4.]])
    >>> w = np.array([0.6, 0.2, 0.15, 0.05])
    >>> utils.resample_equal(x, w)
    array([[ 1.,  1.],
           [ 1.,  1.],
           [ 1.,  1.],
           [ 3.,  3.]])
    
    Notes
    -----
    Implements the systematic resampling method described in `Hol, Schon, and
    Gustafsson (2006) <doi:10.1109/NSSPW.2006.4378824>`_.
    """

    if rstate is None:
        rstate = np.random

    if abs(np.sum(weights) - 1.) > SQRTEPS:  # same tol as in np.random.choice.
        # Guarantee that the weights will sum to 1.
        # warnings.warn("Weights do not sum to 1 and have been renormalized.")
        weights = np.array(weights) / np.sum(weights)

    # Make N subdivisions and choose positions with a consistent random offset.
    nsamples = len(weights)
    positions = (rstate.random() + np.arange(nsamples)) / nsamples

    # Resample the data.
    idx = np.zeros(nsamples, dtype=np.int)
    cumulative_sum = np.cumsum(weights)
    i, j = 0, 0
    while i < nsamples:
        if positions[i] < cumulative_sum[j]:
            idx[i] = j
            i += 1
        else:
            j += 1

    return samples[idx]


class ProgressBar:

    def __init__(self, show=True):
        self.show = show
        if self.show:
            self.progress_bar = tqdm(desc='Iter')

        self.info = dict(beta=None,
                         calls=None,
                         ESS=None,
                         logZ=None,
                         accept=None,
                         N=None,
                         scale=None,
                         corr=None,
                         )

    def update_stats(self, info):
        self.info['beta'] = info.get('beta', self.info['beta'])
        self.info['calls'] = info.get('calls', self.info['calls'])
        self.info['ESS'] = info.get('ESS', self.info['ESS'])
        self.info['logZ'] = info.get('logZ', self.info['logZ'])
        self.info['accept'] = info.get('accept', self.info['accept'])
        self.info['N'] = info.get('N', self.info['N'])
        self.info['scale'] = info.get('scale', self.info['scale'])
        self.info['corr'] = info.get('corr', self.info['corr'])
        if self.show:
            self.progress_bar.set_postfix(ordered_dict=self.info)

    def update_iter(self):
        if self.show:
            self.progress_bar.update(1)

    def close(self):
        if self.show:
            self.progress_bar.close()


class _FunctionWrapper(object):
    r"""
        This is a hack to make the likelihood function pickleable when ``args``
        or ``kwargs`` are also included.

    Parameters
    ----------
    f : (callable)
        Log Probability function.
    args : list
        Extra arguments to be passed into the logprob.
    kwargs : dict
        Extra arguments to be passed into the logprob.

    Returns
    -------
        Log Probability function.
    """

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = [] if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, x):
        return self.f(x, *self.args, **self.kwargs)


def torch_to_numpy(x):
    return x.detach().numpy()


def numpy_to_torch(x):
    return torch.tensor(x, dtype=torch.float32)


def torch_double_to_float(x, warn=True):
    if x.dtype == torch.float64 and warn:
        warnings.warn(f"Float64 data is currently unsupported, casting to Float32. Output will also have type Float32.")
        return x.float()
    elif x.dtype == torch.float32:
        return x
    else:
        raise ValueError(f"Unsupported datatype for input data: {x.dtype}")



from .maf import MAF, RealNVP
import torch
from .tools import torch_double_to_float
import numpy as np
from torch.utils.data import DataLoader, TensorDataset
import copy
import time


def fit(model,
        data,
        context=None,
        validation_data=None,
        validation_context=None,
        validation_split=0.0,
        epochs=20,
        batch_size=100,
        patience=np.inf,
        monitor='val_loss',
        shuffle=True,
        lr=1e-3,
        weight_decay=1e-8,
        clip_grad_norm=1.0,
        l1=None,
        l2=None,
        device='cpu',
        verbose=2):
    r"""
        Method to fit the normalising flow.

    Parameters
    ----------
    data : ``np.ndarray``
        Samples used for training the flow.
    context : ``np.ndarray`` or None
        Additional samples corresponding to conditional  arguments.
    validation_data : ``np.ndarray`` or None
        Samples used for validating the flow.
    validation_context : ``np.ndarray`` or None
        Addiitional samples correspondingg to conditioonal arguments,
        used for validation.
    validation_split : float
        Percentage of ``data`` to be used for validation.
    epochs : int
        Number of training epochs.
    batch_size : int
        Batch size used for training.
    patience : int
        Number of epochs to wait with no improvement until termination.
    monitor : str
        Which loss to monitor for early stopping (e.g. ``val_loss``, ``loss``),
        default is ``val_loss``.
    shuffle : bool
        Shuffle data before training.
    lr : float
        Learning rate.
    weight_decay : float
        Weight decay parameter.
    clip_grad_norm : float
        Clip large gradients (default is ``0``).
    l1 : float or None
        Laplace prior scale for regularisation.
    l2 : float or None
        Gaussian prior scale for regularisation.
    device : str
        Device to use for training, default is ``cpu``.
    verbose : int
        Whether to print all (``2``), some (``1``), or no messages (``0``).

    Returns
    -------
    Training history dictionary.
    """

    optimizer = torch.optim.Adam(model.parameters(), lr, weight_decay=weight_decay)

    if not isinstance(data, torch.Tensor):
        data = torch.tensor(data, dtype=torch.float32)

    if (validation_data is not None) and (not isinstance(validation_data, torch.Tensor)):
        validation_data = torch.tensor(validation_data, dtype=torch.float32)

    if context is not None:
        use_context = True
        if not isinstance(context, torch.Tensor):
            context = torch.tensor(context, dtype=torch.float32)
    else:
        use_context = False

    if (validation_context is not None) and (not isinstance(validation_context, torch.Tensor)):
        validation_context = torch.tensor(validation_context, dtype=torch.float32)

    if validation_data is not None:

        if use_context:
            train_dl = DataLoader(TensorDataset(data, context), batch_size, shuffle)
            val_dl = DataLoader(TensorDataset(validation_data, validation_context), batch_size, shuffle)
        else:
            train_dl = DataLoader(TensorDataset(data), batch_size, shuffle)
            val_dl = DataLoader(TensorDataset(validation_data), batch_size, shuffle)

        validation = True
    else:
        if validation_split > 0.0 and validation_split < 1.0:
            validation = True
            split = int(data.size()[0] * (1. - validation_split))
            if use_context:
                data, validation_data = data[:split], data[split:]
                context, validation_context = context[:split], context[split:]
                train_dl = DataLoader(TensorDataset(data, context), batch_size, shuffle)
                val_dl = DataLoader(TensorDataset(validation_data, validation_context), batch_size, shuffle)
            else:
                data, validation_data = data[:split], data[split:]
                train_dl = DataLoader(TensorDataset(data), batch_size, shuffle)
                val_dl = DataLoader(TensorDataset(validation_data), batch_size, shuffle)
        else:
            validation = False
            if use_context:
                train_dl = DataLoader(TensorDataset(data, context), batch_size, shuffle)
            else:
                train_dl = DataLoader(TensorDataset(data), batch_size, shuffle)

    history = {}  # Collects per-epoch loss
    history['loss'] = []
    history['val_loss'] = []

    if not validation:
        monitor = 'loss'
    best_epoch = 0
    best_loss = np.inf
    best_model = copy.deepcopy(model.state_dict())

    start_time_sec = time.time()

    for epoch in range(epochs):

        # --- TRAIN AND EVALUATE ON TRAINING SET -----------------------------
        model.train()
        train_loss = 0.0

        for batch in train_dl:

            optimizer.zero_grad()

            if use_context:
                x = batch[0].to(device)
                y = batch[1].to(device)
                loss = -model.log_prob(x, y).sum()
            else:
                x = batch[0].to(device)
                loss = -model.log_prob(x).sum()

            if l1 is not None:
                loss -= model.log_prior(scale=l1, type='Laplace')

            if l2 is not None:
                loss -= model.log_prior(scale=l2, type='Gaussian')

            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), clip_grad_norm)
            optimizer.step()

            # train_loss += loss.data.item() * x.size(0)
            train_loss += loss.data.item()

        train_loss = train_loss / len(train_dl.dataset)

        history['loss'].append(train_loss)

        # --- EVALUATE ON VALIDATION SET -------------------------------------
        model.eval()
        if validation:
            val_loss = 0.0

            for batch in val_dl:

                if use_context:
                    x = batch[0].to(device)
                    y = batch[1].to(device)
                    loss = -model.log_prob(x, y).sum()
                else:
                    x = batch[0].to(device)
                    loss = -model.log_prob(x).sum()

                if l1 is not None:
                    loss -= model.log_prior(scale=l1, type='Laplace')

                if l2 is not None:
                    loss -= model.log_prior(scale=l2, type='Gaussian')

                # val_loss += loss.data.item() * x.size(0)
                val_loss += loss.data.item()

            val_loss = val_loss / len(val_dl.dataset)

            history['val_loss'].append(val_loss)

        if verbose > 1:
            try:
                print('Epoch %3d/%3d, train loss: %5.2f, val loss: %5.2f' % \
                      (epoch + 1, epochs, train_loss, val_loss))
            except:
                print('Epoch %3d/%3d, train loss: %5.2f' % \
                      (epoch + 1, epochs, train_loss))

        # Monitor loss
        if history[monitor][-1] < best_loss:
            best_loss = history[monitor][-1]
            best_epoch = epoch
            best_model = copy.deepcopy(model.state_dict())

        if epoch - best_epoch >= patience:
            model.load_state_dict(best_model)
            if verbose > 0:
                print('Finished early after %3d epochs' % (best_epoch))
                print('Best loss achieved %5.2f' % (best_loss))
            break

    # END OF TRAINING LOOP

    if verbose > 0:
        end_time_sec = time.time()
        total_time_sec = end_time_sec - start_time_sec
        time_per_epoch_sec = total_time_sec / epochs
        print()
        print('Time total:     %5.2f sec' % (total_time_sec))
        print('Time per epoch: %5.2f sec' % (time_per_epoch_sec))

    return history


class Flow:
    r"""
    
        Normalising Flow object.

    Parameters
    ----------
    ndim : int
        Number of parameters or dimensions
    flow_config : dict or None
        Configuration of the flow. If ``None`` the default configuration is used ``dict(n_blocks=6,
        hidden_size= 3 * ndim, n_hidden=1, batch_norm=True, activation='relu', input_order='sequential',
        flow_type='maf')``
    train_config : dict or None
        Training confiiguration for the flow. If ``None`` the default configuration is used ``validation_split=0.2,
        epochs=1000, batch_size=nparticles, patience=30, monitor='val_loss', shuffle=True, lr=[1e-2, 1e-3, 1e-4, 1e-5],
        weight_decay=1e-8, clip_grad_norm=1.0, l1=0.2, l2=None, device='cpu', verbose=0``
    """
    def __init__(self, ndim: int, flow_config: dict = None, train_config: dict = None):
        if ndim == 1:
            raise ValueError(f"1D data is not supported. Please provide data with ndim >= 2.")

        self.ndim = ndim
        self.flow_config = flow_config  # TODO do we have to store this?
        self.train_config = train_config

        self.default_config = dict(
            n_blocks=6,
            hidden_size=3 * self.ndim,
            n_hidden=1,
            batch_norm=True,
            activation='relu',
            input_order='sequential',
            flow_type='maf'
        )
        self.flow = self.create(flow_config)

    def validate_config(self, config: dict):
        """
        Raise a ValueError if the flow config dictionary contains invalid inputs.

        Parameters
        ----------
        config: dictionary with key-value pairs to be passed onto constructors of MAF or RealNVP.
        """
        allowed_keys = list(self.default_config.keys())
        for k in config.keys():
            if k not in allowed_keys:
                raise ValueError(f"Unrecognized config key: {k}. Allowed keys are: {allowed_keys}")

    def create(self, config: dict = None):
        """
        Create a normalizing flow based on the desired flow configuration.

        Parameters
        ----------
        config: dictionary with key-value pairs to be passed onto constructors of MAF or RealNVP.

        Returns
        -------
        A MAF or RealNVP object with the desired configuration.
        """
        if config is None:
            config = dict()
        self.validate_config(config)
        config = {**self.default_config, **config}  # Overwrite keys in default_config and add new ones

        if config['flow_type'].lower() == 'maf':
            return MAF(input_size=self.ndim, cond_label_size=None, **config)
        elif config['flow_type'].lower() == 'realnvp':
            return RealNVP(input_size=self.ndim, cond_label_size=None, **config)
        else:
            raise ValueError(f"Unsupported flow type: {config['flow_type']}. Please use one of ['maf', 'realnvp'].")

    def fit(self, x: torch.Tensor):
        """
        Train the normalising flow on the provided data.

        Parameters
        ----------
        x : torch.Tensor
            Training data.
        
        Returns
        -------
        Training history.
        """
        x = torch_double_to_float(x)

        default_train_config = dict(
            validation_split=0.2,
            epochs=1000,
            batch_size=x.shape[0],
            patience=30,
            monitor='val_loss',
            shuffle=True,
            lr=[1e-2, 1e-3, 1e-4, 1e-5],
            weight_decay=1e-8,
            clip_grad_norm=1.0,
            l1=0.2,
            l2=None,
            device='cpu',
            verbose=0
        )
        if self.train_config is None:
            self.train_config = dict()
        train_config = {**default_train_config, **self.train_config}

        history = None
        for lr in train_config['lr']:
            train_config_tmp = copy.deepcopy(train_config)
            del train_config_tmp['lr']
            history = fit(self.flow, x, **train_config_tmp, lr=lr)

        return history

    def forward(self, x: torch.Tensor):
        """
        Forward transformation.
        
        Parameters
        ----------
        x : torch.Tensor
            Samples to transform.
        Returns
        -------
        Tranformed samples.
        """
        x = torch_double_to_float(x)
        return self.flow.forward(x)

    def inverse(self, u: torch.Tensor):
        """
        Inverse transformation.
        
        Parameters
        ----------
        u : torch.Tensor
            Samples to transform.
        Returns
        -------
        Tranformed samples.
        """
        u = torch_double_to_float(u)
        return self.flow.inverse(u)

    def logprob(self, x: torch.Tensor):
        """
        Log-probability of samples.
        
        Parameters
        ----------
        x : torch.Tensor
            Input samples
        Returns
        -------
        Log-probability of samples.
        """
        x = torch_double_to_float(x)
        u, logdetJ = self.flow.forward(x)
        return torch.sum(self.flow.base_dist.log_prob(u) + logdetJ, dim=1)

    def sample(self, size: int = 1):
        """
        Method that generates random samples.

        Parameters
        ----------
        size : int
            Number of samples to generate.
        Returns
        -------
        samples and respective log-probability values. 
        """
        u = torch.randn(size, self.ndim)
        x, logdetJ = self.flow.inverse(u)
        return x, torch.sum(self.flow.base_dist.log_prob(u) + logdetJ, dim=1)



import numpy as np

class Reparameterise:
    """
        Function that reparameterises the model using change-of-variables parameter transformations.
    
    Parameters
    ----------
    ndim : int
        Dimensionality of sampling problem
    bounds : array or list or None
        Parameter bounds
    periodic : list
        List of indeces corresponding to parameters with periodic boundary conditions
    reflective : list
        List of indeces corresponding to parameters with periodic boundary conditions
    scale : bool
        Rescale parameters to zero mean and unit variance (default is true)
    diagonal : bool
        Use diagonal transformation (i.e. ignore covariance) (default is true)
    """
    def __init__(self,
                 ndim,
                 bounds=None,
                 periodic=None,
                 reflective=None,
                 scale=True,
                 diagonal=True
                 ):
                          
        self.ndim = ndim 

        if bounds is None:
            bounds = np.full((self.ndim, 2), np.nan)
        elif len(bounds) == 2 and not np.shape(bounds) == (2,2):
            bounds = np.tile(np.array(bounds).reshape(2,1), self.ndim).T
        self.low = bounds.T[0]
        self.high = bounds.T[1]

        self.periodic = periodic
        self.reflective = reflective

        self.mu = None
        self.sigma = None
        self.cov = None
        self.L = None
        self.Linv = None
        self.logdetL = None
        self.scale = scale
        self.diagonal = diagonal

        self._create_masks()

    def apply_boundary_conditions(self, x):
        """
            Apply boundary conditions (i.e. periodic or reflective) to input.
        
        Parameters
        ----------
        x : array
            Input array
        
        Returns
        -------
        Transformed input
        """
        if (self.periodic is None) and (self.reflective is None):
            return x 
        elif self.periodic is None:
            return self._apply_reflective_boundary_conditions(x)
        elif self.reflective is None:
            return self._apply_periodic_boundary_conditions(x)
        else:
            return self._apply_reflective_boundary_conditions(self._apply_periodic_boundary_conditions(x))

    def _apply_periodic_boundary_conditions(self, x):
        """
            Apply periodic boundary conditions to input.
        
        Parameters
        ----------
        x : array
            Input array
        
        Returns
        -------
        Transformed input.
        """
        if self.periodic is not None:
            x = x.copy()
            for i in self.periodic:
                for j in range(len(x)):
                    while x[j,i] > self.high[i]:
                        x[j,i] = self.low[i] + x[j,i] - self.high[i]
                    while x[j,i] < self.low[i]:
                        x[j,i] = self.high[i] + x[j,i] - self.low[i]
        return x

    def _apply_reflective_boundary_conditions(self, x):
        """
            Apply reflective boundary conditions to input.
        
        Parameters
        ----------
        x : array
            Input array
        
        Returns
        -------
        Transformed input.
        """
        if self.reflective is not None:
            x = x.copy()
            for i in self.reflective:
                for j in range(len(x)):
                    while x[j,i] > self.high[i]:
                        x[j,i] = self.high[i] - x[j,i] + self.high[i]
                    while x[j,i] < self.low[i]:
                        x[j,i] = self.low[i] + self.low[i] - self.x[j,i]

        return x

    def fit(self, x):
        """
            Learn mean and standard deviation using for rescaling.
        
        Parameters
        ----------
        x : array
            Input data used for training.
        """
        u = self._forward(x)
        self.mu = np.mean(u, axis=0)
        if self.diagonal:
            self.sigma = np.std(u, axis=0)
        else:
            self.cov = np.cov(u.T)
            self.L = np.linalg.cholesky(self.cov)
            self.Linv = np.linalg.inv(self.L)
            self.logdetL = np.linalg.slogdet(self.L)[1]

    def forward(self, x):
        """
            Forward transformation (both logit for bounds and affine for all parameters).

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        u : array
            Transformed input data
        """
        u = self._forward(x)
        if self.scale:
            u = self._forward_affine(u)

        return u

    def inverse(self, u):
        """
            Inverse transformation (both logit^-1 for bounds and affine for all parameters).

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        logdetJ : array
            Logarithm of determinant of Jacobian matrix transformation.
        """
        if self.scale:
            x, logdetJ = self._inverse_affine(u)
            x, logdetJ_prime = self._inverse(x)
            logdetJ += logdetJ_prime
        else:
            x, logdetJ = self._inverse(u)

        return x, logdetJ

    def _forward(self, x):
        """
            Forward transformation (obly logit for bounds).

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        u : array
            Transformed input data
        """
        u = np.empty(x.shape)
        u[:,self.mask_none] = self._forward_none(x)
        u[:,self.mask_left] = self._forward_left(x)
        u[:,self.mask_right] = self._forward_right(x)
        u[:,self.mask_both] = self._forward_both(x)

        return u

    def _inverse(self, u):
        """
            Inverse transformation (only logit^-1 for bounds).

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        logdetJ : array
            Logarithm of determinant of Jacobian matrix transformation.
        """
        x = np.empty(u.shape)
        J = np.empty(u.shape)

        x[:, self.mask_none], J[:, self.mask_none] = self._inverse_none(u)
        x[:, self.mask_left], J[:, self.mask_left] = self._inverse_left(u)
        x[:, self.mask_right], J[:, self.mask_right] = self._inverse_right(u)
        x[:, self.mask_both], J[:, self.mask_both] = self._inverse_both(u)

        logdetJ = np.array([np.linalg.slogdet(Ji*np.identity(len(Ji)))[1] for Ji in J])

        return x, logdetJ

    def _forward_affine(self, x):
        """
            Forward affine transformation.

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        Transformed input data
        """
        if self.diagonal:
            return ( x - self.mu ) / self.sigma
        else:
            return np.array([np.dot(self.Linv, xi - self.mu) for xi in x])

    def _inverse_affine(self, u):
        """
            Inverse affine transformation.

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        J : array
            Diagonal of Jacobian matrix.
        """
        if self.diagonal:
            J = self.sigma
            logdetJ = np.linalg.slogdet(J * np.identity(len(J)))[1]
            return self.mu + self.sigma * u, logdetJ * np.ones(len(u))
        else:
            x = self.mu + np.array([np.dot(self.L,ui) for ui in u])
            return x, self.logdetL * np.ones(len(u))

    def _forward_left(self, x):
        """
            Forward transformation for bounded parameters (only low).

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        Transformed input data
        """
        return np.log( x[:,self.mask_left] - self.low[self.mask_left] )

    def _inverse_left(self, u):
        """
            Inverse transformation for bounded parameters (only low).

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        J : array
            Diagonal of Jacobian matrix.
        """
        p = np.exp( u[:,self.mask_left] )

        return p + self.low[self.mask_left], p

    def _forward_right(self, x):
        """
            Forward transformation for bounded parameters (only high).

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        Transformed input data
        """
        return np.log( self.high[self.mask_right] - x[:,self.mask_right] )

    def _inverse_right(self, u):
        """
            Inverse transformation for bounded parameters (only high).

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        J : array
            Diagonal of Jacobian matrix.
        """
        p = np.exp( u[:,self.mask_right] )

        return self.high[self.mask_right] - p, p

    def _forward_both(self, x):
        """
            Forward transformation for bounded parameters (both low and high).

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        Transformed input data
        """
        p = (x[:,self.mask_both] - self.low[self.mask_both]) / (self.high[self.mask_both]-self.low[self.mask_both])

        return np.log( p / ( 1 - p ) )

    def _inverse_both(self, u):
        """
            Inverse transformation for bounded parameters (both low and high).

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        J : array
            Diagonal of Jacobian matrix.
        """
        p = np.exp(-np.logaddexp(0, -u[:,self.mask_both]))

        x = p * (self.high[self.mask_both] - self.low[self.mask_both]) + self.low[self.mask_both]

        J = (self.high[self.mask_both]-self.low[self.mask_both]) * p * (1.0 - p)

        return x, J

    def _forward_none(self, x):
        """
            Forward transformation for unbounded parameters (this does nothing).

        Parameters
        ----------
        x : array
            Input data
        Returns
        -------
        u : array
            Transformed input data
        """
        return x[:,self.mask_none]

    def _inverse_none(self, u):
        """
            Inverse transformation for unbounded parameters (this does nothing).

        Parameters
        ----------
        u : array
            Input data
        Returns
        -------
        x : array
            Transformed input data
        logdetJ : array
            Logarithm of determinant of Jacobian matrix transformation.
        """
        return u[:,self.mask_none], np.ones(u.shape)[:,self.mask_none]

    def _create_masks(self):
        """
            Create parameter masks for bounded parameters
        """

        self.mask_left = np.zeros(self.ndim, dtype=bool)
        self.mask_right = np.zeros(self.ndim, dtype=bool)
        self.mask_both = np.zeros(self.ndim, dtype=bool)
        self.mask_none = np.zeros(self.ndim, dtype=bool)

        # TODO: Do this more elegantly, it's a shame
        for i in range(self.ndim):
            if np.isnan(self.low[i]) and np.isnan(self.high[i]):
                self.mask_none[i] = True
                self.mask_left[i] = False
                self.mask_right[i] = False
                self.mask_both[i] = False
            elif np.isnan(self.low[i]) and not np.isnan(self.high[i]):
                self.mask_none[i] = False
                self.mask_left[i] = False
                self.mask_right[i] = True
                self.mask_both[i] = False
            elif not np.isnan(self.low[i]) and np.isnan(self.high[i]):
                self.mask_none[i] = False
                self.mask_left[i] = True
                self.mask_right[i] = False
                self.mask_both[i] = False
            else:
                self.mask_none[i] = False
                self.mask_left[i] = False
                self.mask_right[i] = False
                self.mask_both[i] = True


import numpy as np
import torch

from .mcmc import PreconditionedMetropolis, Metropolis
from .tools import resample_equal, _FunctionWrapper, torch_to_numpy, numpy_to_torch, get_ESS, ProgressBar
from .scaler import Reparameterise
from .flow import Flow

class Sampler:
    r"""Main Precondioned Monte Carlo sampler class.

    Parameters
    ----------
    nparticles : int
        The total number of particles/walkers to use.
    ndim : int
        The total number of parameters/dimensions.
    loglikelihood : callable
        Function returning the log likelihood of a set
        of parameters.
    logprior : callable
        Function returning the log prior of a set
        of parameters.
    bounds : ``np.ndarray`` or None
        Array of shape ``(ndim, 2)`` holding the boundaries
        of parameters (default is ``bounds=None``). If a
        parameter is unbounded from below, above or both
        please provide ``None`` for the respective boundary.
    threshold : float
        The threshold value for the (normalised) proposal
        scale parameter below which normalising flow
        preconditioning (NFP) is enabled (default is
        ``threshold=1.0``, meaning that NFP is used all
        the time).
    scale : bool
        Scale
    rescale : bool
        Rescale
    diagonal : bool
        Diagonal
    loglikelihood_args : list
        Extra arguments to be passed into the loglikelihood
        (default is ``loglikelihood_args=None``).
    loglikelihood_kwargs : list
        Extra arguments to be passed into the loglikelihood
        (default is ``loglikelihood_kwargs=None``).
    logprior_args : list
        Extra arguments to be passed into the logprior
        (default is ``logprior_args=None``).
    logprior_kwargs : list
        Extra arguments to be passed into the logprior
        (default is ``logprior_kwargs=None``).
    vectorize_likelihood : bool
        Whether or not to vectorize the ``loglikelihood``
        calculation (default is ``vectorize_likelihood=False``).
    vectorize_logprior : bool
        Whether or not to vectorize the ``logprior``
        calculation (default is ``vectorize_prior=False``).
    pool : pool
        Provided ``MPI`` or ``multiprocessing`` pool for
        parallelisation (default is ``pool=None``).
    parallelize_prior : bool
        Whether or not to use the ``pool`` (if provided)
        for the ``logprior`` as well (default is 
        ``parallelize_prior=False``).
    flow_config : dict or ``None``
        Configuration of the normalizing flow (default is
        ``flow_config=None``).
    train_config : dict or ``None``
        Configuration for training the normalizing flow
        (default is ``train_config=None``).
    
    Attributes
    ----------
    results : dict
        Dictionary holding results. Includes the following
        properties: ``iter``, ``samples``, ``posterior_samples``,
        ``logw``, ``logl``, ``logz``, ``ess``, ``ncall``, 
        ``beta``, ``accept``, ``scale``, and ``steps``.
    """

    def __init__(self,
                 nparticles,
                 ndim,
                 loglikelihood,
                 logprior,
                 bounds=None,
                 periodic=None,
                 reflective=None,
                 threshold=1.0,
                 scale=True,
                 rescale=False,
                 diagonal=True,
                 loglikelihood_args=None,
                 loglikelihood_kwargs=None,
                 logprior_args=None,
                 logprior_kwargs=None,
                 vectorize_likelihood=False,
                 vectorize_prior=False,
                 pool=None,
                 parallelize_prior=False,
                 flow_config=None,
                 train_config=None,
                 random_state: int = None
                 ):
        if random_state is not None:
            np.random.seed(random_state)
            torch.manual_seed(random_state)

        self.nwalkers = nparticles
        self.ndim = ndim
        
        # Distributions
        self.loglikelihood = _FunctionWrapper(loglikelihood,
                                              loglikelihood_args, 
                                              loglikelihood_kwargs)
        self.logprior = _FunctionWrapper(logprior, 
                                         logprior_args, 
                                         logprior_kwargs)

        # Sampling
        self.ncall = 0
        self.t = 0
        self.beta = 0.0
        self.logw = 0.0
        self.sum_logw = 0.0
        self.logz = 0.0

        # Results
        self.saved_iter = []
        self.saved_samples = []
        self.saved_posterior_samples = []
        self.saved_posterior_logl = []
        self.saved_posterior_logp = []
        self.saved_logl = []
        self.saved_logp = []
        self.saved_logw = []
        self.saved_logz = []
        self.saved_ess = []
        self.saved_ncall = []
        self.saved_beta = []
        self.saved_accept = []
        self.saved_scale = []
        self.saved_steps = []

        # State variables
        self.u = None
        self.x = None
        self.J = None
        self.L = None
        self.P = None

        # Parallelism
        self.pool = pool
        if pool is None:
            self.distribute = map
        else:
            self.distribute = pool.map
        self.vectorize_likelihood = vectorize_likelihood
        self.vectorize_prior = vectorize_prior
        self.parallelize_prior = parallelize_prior

        # Flow
        self.flow = Flow(self.ndim, flow_config, train_config)
        self.threshold = threshold
        self.use_flow = False
        
        # Scaler
        self.scaler = Reparameterise(self.ndim, bounds, periodic, reflective, scale, diagonal)
        self.rescale = rescale

        # MCMC parameters
        self.ideal_scale = 2.38 / np.sqrt(ndim)
        self.scale = 2.38 / np.sqrt(ndim)
        self.accept = 0.234
        self.target_accept = 0.234

    def run(self,
            x0,
            ess = 0.95,
            gamma = 0.75,
            nmin = None,
            nmax = None,
            progress=True
            ):
        r"""Method that runs Preconditioned Monte Carlo.

        Parameters
        ----------
        x0 : ``np.ndarray``
            Array holding the initial positions of the particles. The initial
            positions must be sampled from the prior distribution.
        ess : float
            The effective sample size maintained during the run (default is
            `ess=0.95`).
        gamma : float
            Threshold for the correlation coefficient that is
            used to adaptively determine the number of MCMC
            steps (default is ``gamma=0.75``).
        nmin : int or None
            The minimum number of MCMC steps per iteration (default is ``nmin = ndim // 2``).
        nmax : int or None
            The maximum number of MCMC steps per iteration  (default is ``nmin = int(10 * ndim)``).
        progress : bool
            Whether or not to print progress bar (default is ``progress=True``).        
        """

        # Run parameters
        self.ess = ess
        self.gamma = gamma
        if nmin is None:
            self.nmin = self.ndim // 2
        else:
            self.nmin = int(nmin)
        if nmax is None:
            self.nmax = int(10 * self.ndim)
        else:
            self.nmax = int(nmax)
        self.progress = progress

        # Set state parameters
        self.x = np.copy(x0)
        self.scaler.fit(self.x)
        self.u = self.scaler.forward(self.x)
        self.J = self.scaler.inverse(self.u)[1]
        self.P = self._logprior(self.x)
        self.L = self._loglike(self.x)
        self.ncall += len(self.x)

        # Pre-train flow if required
        if self.threshold >= 1.0:
            self.use_flow = True
            self._train(self.u)

        # Save state
        self.saved_samples.append(self.x)
        self.saved_logl.append(self.L)
        self.saved_logp.append(self.P)
        self.saved_iter.append(self.t)
        self.saved_beta.append(self.beta)
        self.saved_logw.append(np.zeros(self.nwalkers))
        self.saved_logz.append(self.logz)
        self.saved_ess.append(self.ess)
        self.saved_ncall.append(self.ncall)
        self.saved_accept.append(self.accept)
        self.saved_scale.append(self.scale/self.ideal_scale)
        self.saved_steps.append(0)

        # Initialise progress bar
        self.pbar = ProgressBar(self.progress)
        self.pbar.update_stats(dict(beta=self.beta,
                                    calls=self.ncall,
                                    ESS=self.ess,
                                    logZ=self.logz,
                                    accept=0.234,
                                    N=0,
                                    scale=1.0))
        
        # Run Sequential Monte Carlo
        while 1.0 - self.beta >= 1e-4:

            # Choose next beta based on CV of weights
            self._update_beta()

            # Resample x_prev, w
            self.u, self.x, self.J, self.L, self.P = self._resample(self.u,
                                                                    self.x,
                                                                    self.J,
                                                                    self.L,
                                                                    self.P)

            # Evolve particles using MCMC
            self.u, self.x, self.J, self.L, self.P = self._mutate(self.u,
                                                                  self.x,
                                                                  self.J,
                                                                  self.L,
                                                                  self.P)

            # Rescale parameters
            if self.rescale:
                self.scaler.fit(self.x)
                self.u = self.scaler.forward(self.x)
                self.J = self.scaler.inverse(self.u)[1]

            # Train Precondiotoner
            self._train(self.u)

        self.saved_posterior_samples.append(self.x.copy())
        self.saved_posterior_logl.append(self.L.copy())
        self.saved_posterior_logp.append(self.P.copy())
        
        self.pbar.close()

    
    def add_samples(self,
                    N=1000,
                    retrain=False, 
                    progress=True
                    ):
        r"""Method that generates additional samples at the end of the run

        Parameters
        ----------
        N : int
            The number of additional samples (default is ``N=1000``).
        retrain : bool
            Whether or not to retrain the normalising flow preconditioner
            between iterations (default is ``retrain=False``).
        progress : bool
            Whether or not to print progress bar (default is ``progress=True``).
        """
        self.progress = progress

        self.pbar = ProgressBar(self.progress)
        self.pbar.update_stats(dict(beta=self.beta,
                                    calls=self.ncall,
                                    ESS=self.ess,
                                    logZ=self.logz,
                                    accept=0,
                                    N=0,
                                    scale=0))

        iterations = int(np.ceil(N/len(self.u)))
        for _ in range(iterations):
            self.u, self.x, self.J, self.L, self.P = self._mutate(self.u,
                                                                  self.x,
                                                                  self.J,
                                                                  self.L,
                                                                  self.P)
            if retrain:
                self._train(self.u)
            self.saved_posterior_samples.append(self.x)
            self.saved_posterior_logl.append(self.L)
            self.saved_posterior_logp.append(self.P)
            self.pbar.update_iter()

        self.pbar.close()
        #self.u = np.tile(self.u.T, multiply).T

    
    def _mutate(self, u, x, J, L, P):

        state_dict = dict(u=u.copy(),
                          x=x.copy(),
                          J=J.copy(),
                          L=L.copy(),
                          P=P.copy(),
                          beta=self.beta)

        function_dict = dict(loglike=self._loglike,
                             logprior=self._logprior,
                             scaler=self.scaler,
                             flow=self.flow)

        option_dict = dict(nmin=self.nmin,
                           nmax=self.nmax,
                           corr_threshold=self.gamma,
                           sigma=self.scale,
                           progress_bar=self.pbar)

        if self.use_flow:
            results = PreconditionedMetropolis(state_dict,
                                               function_dict, 
                                               option_dict)
        else:
            results = Metropolis(state_dict,
                                 function_dict,
                                 option_dict)
            
        u = results.get('u').copy()
        x = results.get('x').copy()
        J = results.get('J').copy()
        L = results.get('L').copy()
        P = results.get('P').copy()

        self.scale = results.get('scale')
        self.Nsteps = results.get('steps')
        self.accept = results.get('accept')
        
        self.ncall += self.Nsteps * len(x)

        self.saved_ncall.append(self.ncall)
        self.saved_accept.append(self.accept)
        self.saved_scale.append(self.scale/self.ideal_scale)
        self.saved_steps.append(self.Nsteps)

        return u, x, J, L, P


    def _train(self, u):
        if (self.scale < self.threshold * self.ideal_scale and self.t > 1) or self.use_flow:
            y = np.copy(u)
            np.random.shuffle(y)
            self.flow.fit(numpy_to_torch(y))
            self.use_flow = True
        else:
            pass


    def _resample(self, u, x, J, L, P):
        self.saved_samples.append(x)
        self.saved_logl.append(L)
        self.saved_logp.append(P)
        w = np.exp(self.logw-np.max(self.logw))
        w /= np.sum(w)

        assert np.any(~np.isnan(self.logw))
        assert np.any(np.isfinite(self.logw))
        assert np.any(~np.isnan(w))
        assert np.any(np.isfinite(w))

        try:
            idx = resample_equal(np.arange(len(u)), w)
        except:
            idx = np.random.choice(np.arange(len(u)), p=w, size=len(w))
        self.logw = 0.0

        return u[idx], x[idx], J[idx], L[idx], P[idx]

    
    def _update_beta(self):

        # Update iteration index
        self.t += 1
        self.saved_iter.append(self.t)
        self.pbar.update_iter()

        beta_prev = np.copy(self.beta)
        beta_max = 1.0
        beta_min = np.copy(beta_prev)
        self.logw_prev = np.copy(self.logw)

        while True:

            beta = (beta_max + beta_min) * 0.5
            self.logw = self.logw_prev + self.L * (beta - beta_prev)
            self.ess_est = get_ESS(self.logw)

            if len(self.saved_beta) > 1:
                dbeta = self.saved_beta[-1] - self.saved_beta[-2]

                if 1.0 - beta < dbeta * 0.1:
                    beta = 1.0

            if (np.abs(self.ess_est-self.ess) < min(0.001 * self.ess, 0.001) or beta == 1.0):
                self.saved_beta.append(beta)
                self.saved_logw.append(self.logw)
                self.sum_logw += self.logw
                self.saved_ess.append(self.ess_est)
                self.beta = beta
                self.pbar.update_stats(dict(beta=self.beta, ESS=self.ess_est))
                # Update evidence 
                self.logz += np.mean(self.logw)
                self.saved_logz.append(self.logz)
                self.pbar.update_stats(dict(logZ=self.logz))
                break

            elif self.ess_est < self.ess:
                beta_max = beta 
            else:
                beta_min = beta
        

    def _logprior(self, x):
        if self.vectorize_prior:
            return self.logprior(x)
        elif self.parallelize_prior and self.pool is not None:
            return np.array(list(self.distribute(self.logprior, x)))
        else:
            return np.array(list(map(self.logprior, x)))


    def _loglike(self, x):
        if self.vectorize_likelihood:
            return self.loglikelihood(x)
        elif self.pool is not None:
            return np.array(list(self.distribute(self.loglikelihood, x)))
        else:
            return np.array(list(map(self.loglikelihood, x)))


    def __getstate__(self):
        """Get state information for pickling."""

        state = self.__dict__.copy()

        try:
            #remove random module
            #del state['rstate']

            # deal with pool
            if state['pool'] is not None:
                del state['pool']  # remove pool
                del state['distribute']  # remove `pool.map` function hook
        except:
            pass

        return state

    @property
    def results(self):
        
        results = {
            'samples' : np.vstack(self.saved_posterior_samples),
            'loglikelihood' : np.hstack(self.saved_posterior_logl),
            'logprior' : np.hstack(self.saved_posterior_logp),
            'logz' : np.array(self.saved_logz),
            'iter' : np.array(self.saved_iter),
            'x' : np.array(self.saved_samples),
            'logl' : np.array(self.saved_logl),
            'logp' : np.array(self.saved_logp),
            'logw' : np.array(self.saved_logw),
            'ess' : np.array(self.saved_ess),
            'ncall' : np.array(self.saved_ncall),
            'beta' : np.array(self.saved_beta),
            'accept' : np.array(self.saved_accept),
            'scale' : np.array(self.saved_scale),
            'steps' : np.array(self.saved_steps)
        }

        return results

    def bridge_sampling(self, tolerance=1e-10, maxiter=1000, thin=1):

        x = self.results.get("samples")[::thin]
        l = self.results.get("loglikelihood")[::thin]
        p = self.results.get("logprior")[::thin]
        
        N1 = len(x)
        N2 = len(x)

        s1 = N1 / (N1+N2)
        s2 = N2 / (N1+N2)

        import torch

        u_prop, logg_i = self.flow.sample(size=N2)
        x_prop, J_prop = self.scaler.inverse(torch_to_numpy(u_prop))
        u_prop = torch_to_numpy(u_prop)
        logg_i = torch_to_numpy(logg_i) + J_prop

        u = self.scaler.forward(x)
        x, J = self.scaler.inverse(u)
        logg_j = torch_to_numpy(self.flow.logprob(numpy_to_torch(u))) - J

        logp_i = self._loglike(x_prop) + self._logprior(x_prop) - J_prop
        logp_j = p + l

        logl1j = logp_j - logg_j
        logl2i = logp_i - logg_i

        lstar = max(np.max(logl1j), np.max(logl2i))

        l1j = np.exp(logl1j - lstar)
        l2i = np.exp(logl2i - lstar)

        r = 1.0
        r0 = 0.0
        cnt = 1
        while np.abs(r-r0) > tolerance or cnt <= maxiter:
            r0 = r
            A = np.mean(l2i / (s1 * l2i + s2 * r0))
            B = np.mean(1.0 / (s1 * l1j + s2 * r0))
            r = A / B
            cnt += 1

        return np.log(r) + lstar